import { FileHandle, open } from 'fs/promises'

type Pixel = '.' | '#'
type Computer = {
  state: { x: number }
  cycle: number
  crt: Pixel[][]
}

const matchAddCommand = (line: string) => line.match(/^addx (.+)$/)
const matchNoopCommand = (line: string) => line.match(/^noop$/)

function getCRTWidth(computer: Computer) {
  return computer.crt[0].length
}

function getCRTPosition(computer: Computer) {
  return computer.cycle % getCRTWidth(computer)
}

function shouldLightPixel(computer: Computer): boolean {
  const crtPosition = getCRTPosition(computer)
  return crtPosition >= computer.state.x && crtPosition <= computer.state.x + 2
}

function syncCrtPixel(computer: Computer) {
  const crtWidth = getCRTWidth(computer)
  const cycleNumber = computer.cycle
  const cycleRow = Math.floor(cycleNumber / crtWidth)
  const cycleCol = cycleNumber - cycleRow * crtWidth

  if (shouldLightPixel(computer)) {
    // console.log(`cycle: ${computer.cycle} x: ${computer.state.x} row: ${cycleRow} col: ${cycleCol}`)
    computer.crt[cycleRow][cycleCol - 1] = '#'
  }
}

function tick(computer: Computer, updateState?: Computer['state']) {
  computer.cycle += 1

  syncCrtPixel(computer)

  if (updateState) {
    computer.state = updateState
  }
}

async function solve(file: FileHandle): Promise<void> {
  const computer: Computer = {
    state: { x: 1 },
    crt: Array(6)
      .fill([])
      .map(() => Array(40).fill('.')),
    cycle: 0,
  }

  for await (const line of file.readLines()) {
    const matchAdd = matchAddCommand(line)

    if (matchAdd) {
      const x = +matchAdd[1]
      tick(computer)
      tick(computer, { x: computer.state.x + x })
    } else if (matchNoopCommand(line)) {
      tick(computer)
    } else {
      throw new Error(`Unknown command: ${line}`)
    }
  }

  console.log(computer.crt.map((row) => row.join('')).join('\n'))
}

async function run() {
  const input = process.argv.slice(2)[0]
  const fd = await open(input)
  await solve(fd)
}

run()
