import { FileHandle, open } from 'fs/promises'

type MonkeyLines = [string, string, string, string, string, string]
type Monkey = {
  order: number
  items: number[]
  inspect: (monkeys: Monkey[]) => void
  inspected: number
  testDivisor: number
}

const matchMonkeySearch = /^Monkey (\d+):$/
const matchItemsSearch = /^\s+Starting items: (.+)$/
const matchOperationSearch = /^\s+Operation: new = (old|\d+) (.) (old|\d+)$/
const matchTestSearch = /^\s+Test: divisible by (\d+)$/
const matchTestTrueSearch = /^\s+If true: throw to monkey (\d+)$/
const matchTestFalseSearch = /^\s+If false: throw to monkey (\d+)$/

function parseMonkey(monkeyLines: MonkeyLines): Monkey {
  const matchMonkey = matchMonkeySearch.exec(monkeyLines[0])
  const matchItems = matchItemsSearch.exec(monkeyLines[1])
  const matchOperation = matchOperationSearch.exec(monkeyLines[2])
  const matchTest = matchTestSearch.exec(monkeyLines[3])
  const matchTestTrue = matchTestTrueSearch.exec(monkeyLines[4])
  const matchTestFalse = matchTestFalseSearch.exec(monkeyLines[5])

  if (
    !matchMonkey ||
    !matchItems ||
    !matchOperation ||
    !matchTest ||
    !matchTestTrue ||
    !matchTestFalse
  ) {
    throw new Error(`invalid monkey: ${monkeyLines}`)
  }

  const operation = (item: number): number => {
    const x = matchOperation[1] === 'old' ? item : +matchOperation[1]
    const y = matchOperation[3] === 'old' ? item : +matchOperation[3]

    if (matchOperation[2] === '+') {
      return x + y
    } else {
      return x * y
    }
  }

  const test = (item: number, divisor: number): boolean => {
    return item % divisor === 0
  }

  const monkey: Monkey = {
    order: +matchMonkey[1],
    items: matchItems[1].split(',').map((x) => +x),
    inspect: (monkeys: Monkey[]) => {
      const lcm = monkeys.reduce((total, one) => total * one.testDivisor, 1)
      while (monkey.items.length) {
        const item = monkey.items.shift() as number
        const worry = operation(item) % lcm
        const throwTo = test(worry, monkey.testDivisor) ? +matchTestTrue[1] : +matchTestFalse[1]
        if (monkeys[throwTo]) {
          monkey.inspected += 1
          monkeys[throwTo].items.push(worry)
        } else {
          console.log(`invalid throw: from ${monkey.order} to ${throwTo}`)
        }
      }
    },
    testDivisor: +matchTest[1],
    inspected: 0,
  }

  return monkey
}

function monkeyBusiness(monkeys: Monkey[]): number {
  const inspected = monkeys.map((monkey) => monkey.inspected).sort((a, b) => b - a)
  return inspected[0] * inspected[1]
}

function monkeyRound(monkeys: Monkey[]) {
  monkeys.forEach((monkey) => {
    monkey.inspect(monkeys)
  })
}

function monkeyRounds(number: number, monkeys: Monkey[]) {
  for (let i = 0; i < number; i++) {
    monkeyRound(monkeys)
  }
}

async function solve(file: FileHandle): Promise<number> {
  const monkeys: Monkey[] = []
  const monkeyLines: MonkeyLines = Array(0).fill('') as MonkeyLines

  for await (const line of file.readLines()) {
    if (!/^\s*$/.test(line)) {
      if (monkeyLines.length !== 6) {
        monkeyLines.push(line)
      }

      if (monkeyLines.length === 6) {
        monkeys.push(parseMonkey(monkeyLines))
        monkeyLines.splice(0, monkeyLines.length)
      }
    }
  }

  monkeyRounds(10000, monkeys)

  return monkeyBusiness(monkeys)
}

async function run() {
  const input = process.argv.slice(2)[0]
  const fd = await open(input)
  const solution = await solve(fd)

  console.log(`solution: ${solution}`)
}

run()
