import { FileHandle, open } from 'fs/promises'

type Point = [number, number]

const alpha = 'abcdefghijklmnopqrstuvwxyz'.split('')
const heights = Object.fromEntries(alpha.map((x, i) => [x, i]))
const markerDestination = 'E'
const markerStart = 'S'
const markerLowest = 'a'

heights['S'] = heights['a']
heights['E'] = heights['z']

function canClimb(from: string, to: string): boolean {
  return heights[from] + 1 >= heights[to]
}

function canMoveTo(from: Point, to: Point, map: string[][]): boolean {
  const [x, y] = to
  if (x >= 0 && x < map.length && y >= 0 && y < map[x].length) {
    return canClimb(map[from[0]][from[1]], map[x][y])
  }
  return false
}

function hasVisited(at: Point, visited: Point[]): boolean {
  return visited.some((point) => point[0] === at[0] && point[1] === at[1])
}

function breadthFirstSearch(map: string[][], start: Point, end: Point): number {
  let count = 0
  const queue: [Point, number][] = [[start, count]]
  const visited: Point[] = [start]

  while (queue.length > 0) {
    const [[x, y], depth] = queue.shift() as [Point, number]

    if (x === end[0] && y === end[1]) {
      count = depth
      continue
    }

    const neighbors: Point[] = [
      [x - 1, y],
      [x + 1, y],
      [x, y - 1],
      [x, y + 1],
    ]

    for (const neighbor of neighbors) {
      if (canMoveTo([x, y], neighbor, map) && !hasVisited(neighbor, visited)) {
        queue.push([neighbor, depth + 1])
        visited.push(neighbor)
      }
    }
  }

  return count
}

async function solve(file: FileHandle): Promise<number | undefined> {
  const map: string[][] = []
  const start: Point = [0, 0]
  const end: Point = [0, 0]
  const possibleStarts: Point[] = [start]

  for await (const line of file.readLines()) {
    map.push(line.split(''))

    const findStart = line.indexOf(markerStart)
    const findEnd = line.indexOf(markerDestination)
    const row = map.length - 1

    map[row].forEach((x, i) => {
      if (x === markerLowest) {
        possibleStarts.push([row, i])
      }
    })

    if (findStart >= 0) {
      start[0] = map.length - 1
      start[1] = findStart
    }

    if (findEnd >= 0) {
      end[0] = map.length - 1
      end[1] = findEnd
    }
  }

  return possibleStarts
    .map((start) => breadthFirstSearch(map, start, end))
    .filter((x) => x > 0)
    .sort((a, b) => a - b)[0]
}

async function run() {
  const input = process.argv.slice(2)[0]
  const fd = await open(input)
  const solution = await solve(fd)

  console.log(`solution: ${solution}`)
}

run()
