import { FileHandle, open } from 'fs/promises'

type Packet = number | Packet[]
type PacketCompare = 1 | -1 | 0

function comparePackets(first: Packet[], second: Packet[]): PacketCompare {
  for (let i = 0; i < first.length; i++) {
    const firstValue = first[i]
    const secondValue = second[i]

    if (secondValue === undefined) {
      return 1
    }

    if (typeof firstValue === 'number' && typeof secondValue === 'number') {
      if (firstValue !== secondValue) {
        return firstValue < secondValue ? -1 : 1
      }
    } else {
      let compare: PacketCompare = 0
      compare = comparePackets(
        Array.isArray(firstValue) ? firstValue : [firstValue],
        Array.isArray(secondValue) ? secondValue : [secondValue]
      )

      if (compare !== 0) {
        return compare
      }
    }
  }

  if (first.length === second.length) {
    return 0
  }

  return first.length < second.length ? -1 : 1
}

function parsePacket(line: string): Packet[] {
  const packet: Packet = []
  const stack: Packet[] = [packet]
  let bracketDepth = 0
  let number: number | null = null

  line.split('').forEach((char) => {
    const latestStack = stack[stack.length - 1] as Packet[]

    if (/[[\],]/.test(char) && number !== null) {
      latestStack.push(number)
      number = null
    }

    if (char === '[') {
      if (bracketDepth) {
        const newPacket = [] as Packet
        const latestPacket = stack[stack.length - 1] as Packet[]
        latestPacket.push(newPacket)
        stack.push(newPacket)
      }
      bracketDepth++
    } else if (char === ']') {
      bracketDepth--
      stack.pop()
    } else if (/\d+/g.test(char)) {
      const integer = parseInt(char)
      number = number !== null ? number * 10 + integer : integer
    }
  })

  return packet
}

async function solvePart1(file: FileHandle): Promise<number | undefined> {
  const packets: string[] = []
  const packetPairs: [string, string][] = []

  for await (const line of file.readLines()) {
    if (/^[^\s]+$/.test(line)) {
      packets.push(line)
    }
  }

  while (packets.length) {
    const left = packets.shift()
    const right = packets.shift()

    if (left && right) {
      packetPairs.push([left, right])
    } else {
      throw new Error('invalid packet pair')
    }
  }

  return packetPairs.reduce((acc, [left, right], index) => {
    const leftPacket = parsePacket(left)
    const rightPacket = parsePacket(right)

    return comparePackets(leftPacket, rightPacket) === -1 ? acc + index + 1 : acc
  }, 0)
}

async function solvePart2(file: FileHandle): Promise<number | undefined> {
  const packets: string[] = []
  const decoderPackets: [string, string] = ['[[2]]', '[[6]]']
  const packetPairs: [string, string][] = [decoderPackets]

  for await (const line of file.readLines()) {
    if (/^[^\s]+$/.test(line)) {
      packets.push(line)
    }
  }

  while (packets.length) {
    const left = packets.shift()
    const right = packets.shift()

    if (left && right) {
      packetPairs.push([left, right])
    } else {
      throw new Error('invalid packet pair')
    }
  }

  const sorted = packetPairs
    .flatMap((pair) => [parsePacket(pair[0]), parsePacket(pair[1])])
    .sort((left, right) => {
      return comparePackets(left, right)
    })

  return decoderPackets.reduce((total, packet) => {
    return total * (sorted.findIndex((find) => JSON.stringify(find) === packet) + 1)
  }, 1)
}

async function run() {
  const input = process.argv.slice(2)[0]
  const fd = await open(input)
  const solution1 = await solvePart1(fd)
  console.log(`part1: ${solution1}`)

  const fd2 = await open(input)
  const solution2 = await solvePart2(fd2)
  console.log(`part2: ${solution2}`)
}

run()
