import { readFileSync } from 'fs'

const CaveSpaces = {
  rock: '#',
  empty: '.',
  sand: 'o',
  source: '+',
  infinity: '∞',
} as const

type CaveSpace = typeof CaveSpaces[keyof typeof CaveSpaces]
type CavePoint = [number, number]
type Cave = {
  spaces: CaveSpace[][]
  source: CavePoint
  dimensions: {
    top: number
    bottom: number
    left: number
    right: number
  }
  floor: CaveSpace
  sands: CavePoint[]
}

function printCave(cave: Cave) {
  console.log(cave.spaces.map((row) => row.join('')).join('\n'))
}

function getCavePoint(cave: Cave, [x, y]: CavePoint): CaveSpace {
  if (x - cave.dimensions.left < 0) {
    expandCaveLeft(cave)
  } else if (x - cave.dimensions.left > cave.dimensions.right - cave.dimensions.left) {
    expandCaveRight(cave)
  }

  const space = cave.spaces?.[y - cave.dimensions.top]?.[x - cave.dimensions.left]
  return space ?? CaveSpaces.infinity
}

function setCavePoint(cave: Cave, [x, y]: CavePoint, space: CaveSpace): CavePoint {
  cave.spaces[y - cave.dimensions.top][x - cave.dimensions.left] = space
  return [x, y]
}

function makeCave(paths: number[][][], source: [number, number], floor: CaveSpace): Cave {
  const cave: Cave = {
    spaces: [],
    dimensions: {
      left: source[0],
      right: source[0],
      top: source[1],
      bottom: source[1] + 2,
    },
    source,
    floor,
    sands: [],
  }

  cave.dimensions = paths.reduce((dimensions, path) => {
    const xs = path.map(([x]) => x)
    const ys = path.map(([, y]) => y)
    const minX = Math.min(...xs)
    const maxX = Math.max(...xs)
    const maxY = Math.max(...ys)

    return {
      ...dimensions,
      left: Math.min(dimensions.left, minX - 1),
      right: Math.max(dimensions.right, maxX + 1),
      bottom: Math.max(dimensions.bottom, maxY + 2),
    }
  }, cave.dimensions)

  cave.spaces = Array(cave.dimensions.bottom - cave.dimensions.top + 1)
    .fill('')
    .map(() => Array(cave.dimensions.right - cave.dimensions.left + 1).fill(CaveSpaces.empty))

  paths.forEach((path) => plotRockPaths(cave, path))

  setCavePoint(cave, cave.source, CaveSpaces.source)
  makeCaveFloor(cave)

  return cave
}

function expandCaveLeft(cave: Cave) {
  if (cave.dimensions.left > 0) {
    cave.spaces.forEach((row) => {
      row.unshift(CaveSpaces.empty)
    })
    cave.dimensions.left += -1

    makeCaveFloor(cave)
  }
}

function expandCaveRight(cave: Cave) {
  cave.spaces.forEach((row) => {
    row.push(CaveSpaces.empty)
  })

  cave.dimensions.right += 1

  makeCaveFloor(cave)
}

function makeCaveFloor(cave: Cave) {
  const bottom = cave.dimensions.bottom
  const floor = cave.spaces[bottom]
  floor.forEach((s, x) => {
    floor[x] = s === CaveSpaces.empty ? cave.floor : s
  })
}

function isSource(cave: Cave, [x, y]: CavePoint): boolean {
  return x === cave.source[0] && y === cave.source[1]
}

function isSandInRow(cave: Cave, row: number): boolean {
  return cave.spaces.slice(row)[0].some((space) => space === CaveSpaces.sand)
}

function plotRockPaths(cave: Cave, path: number[][]) {
  path.reduce((from, to) => {
    const [toX, toY] = to

    if (from.length == 2 && to.length == 2) {
      const [fromX, fromY] = from

      if (fromY === toY) {
        for (let x = Math.min(fromX, toX); x <= Math.max(fromX, toX); x++) {
          setCavePoint(cave, [x, toY], CaveSpaces.rock)
        }
      } else {
        for (let y = Math.min(fromY, toY); y <= Math.max(fromY, toY); y++) {
          setCavePoint(cave, [fromX, y], CaveSpaces.rock)
        }
      }
    }
    return to
  }, [])
}

function dropSandOneStep(sand: CavePoint, cave: Cave): boolean {
  const [x, y] = sand
  const tryMoves: CavePoint[] = [
    [x, y + 1],
    [x - 1, y + 1],
    [x + 1, y + 1],
  ]

  const moved = tryMoves.find((move) => {
    if (getCavePoint(cave, move) === CaveSpaces.empty) {
      setCavePoint(cave, sand, isSource(cave, sand) ? CaveSpaces.source : CaveSpaces.empty)
      setCavePoint(cave, move, CaveSpaces.sand)
      sand.splice(0, 2, ...move)
      return true
    }
    return false
  })

  return moved !== undefined
}

function dropSand(cave: Cave) {
  const sand = [...cave.source] as CavePoint
  cave.sands.push(sand)
  setCavePoint(cave, sand, CaveSpaces.sand)

  while (dropSandOneStep(sand, cave)) {
    // keep dropping
  }
}

function solvePart1(lines: string[]): number {
  const paths = lines.map((line) => line.split(' -> ').map((point) => point.split(',').map(Number)))
  const cave = makeCave(paths, [500, 0], CaveSpaces.empty)

  while (!isSandInRow(cave, -1)) {
    dropSand(cave)
  }

  if (process.env.DEBUG) {
    printCave(cave)
  }

  return cave.sands.length - 1
}

function solvePart2(lines: string[]): number {
  const paths = lines.map((line) => line.split(' -> ').map((point) => point.split(',').map(Number)))

  const cave = makeCave(paths, [500, 0], CaveSpaces.rock)

  while (!isSandInRow(cave, 0)) {
    dropSand(cave)
  }

  if (process.env.DEBUG) {
    printCave(cave)
  }

  return cave.sands.length
}

function run() {
  const input = process.argv.slice(2)[0]
  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)
  const solution1 = solvePart1(lines)
  console.log(`part1: ${solution1}`)

  const solution2 = solvePart2(lines)
  console.log(`part2: ${solution2}`)
}

run()
