import { readFileSync } from 'fs'

type Point = { x: number; y: number }
type Sensor = Point & { beacon: Point; distance: number }

const parser = /Sensor at x=(-?\d+), y=(-?\d+): closest beacon is at x=(-?\d+), y=(-?\d+)/

function manHattanDistance(a: Point, b: Point) {
  return Math.abs(a.x - b.x) + Math.abs(a.y - b.y)
}

function isWithinSensorRange({ x, y, distance }: Sensor, point: Point) {
  return manHattanDistance({ x, y }, point) <= distance
}

function isUncovered(sensors: Sensor[], point: Point): boolean {
  return sensors.every((sensor) => {
    return !isWithinSensorRange(sensor, point)
  })
}

function findCoverageForRow(sensors: Sensor[], row: number) {
  const coverage = new Set<number>()
  const numSensors = sensors.reduce((count, s) => count + (s.y === row ? 1 : 0), 0)
  const numBeacons = sensors
    .map((s) => s.beacon)
    .reduce((list, b) => {
      list[b.y] = list[b.y] || new Set()
      list[b.y].add(b.x)
      return list
    }, {} as { [key: number]: Set<number> })?.[row].size

  sensors.forEach((sensor) => {
    const { x, distance } = sensor

    for (let i = x - distance; i <= x + distance; i++) {
      if (isWithinSensorRange(sensor, { x: i, y: row })) {
        coverage.add(i)
      }
    }
  })

  return coverage.size - numBeacons - numSensors
}

function withinBounds(x: number, y: number, high: number, low: number) {
  return x <= high && x >= low && y >= low && y <= high
}

function searchPerimeter(sensors: Sensor[], low: number, high: number): Point | undefined {
  let x = 0
  let y = 0

  const found = sensors.some((sensor) => {
    const perimeter = 1
    const lowX = sensor.x - sensor.distance - perimeter
    const lowY = sensor.y - sensor.distance - perimeter
    const highX = sensor.x + sensor.distance + perimeter
    const highY = sensor.y + sensor.distance + perimeter

    const ranges = [
      {
        start: () => [lowX, sensor.y],
        end: (x: number, y: number) => y <= lowY && withinBounds(x, y, high, low),
        step: (x: number, y: number) => [x - 1, y - 1],
      },
      {
        start: () => [lowX, sensor.y],
        end: (x: number, y: number) => y <= highY && withinBounds(x, y, high, low),
        step: (x: number, y: number) => [x - 1, y + 1],
      },
      {
        start: () => [sensor.x, lowY],
        end: (x: number, y: number) => x <= highX && withinBounds(x, y, high, low),
        step: (x: number, y: number) => [x + 1, y + 1],
      },
      {
        start: () => [sensor.x, highY],
        end: (x: number, y: number) => x <= lowX && withinBounds(x, y, high, low),
        step: (x: number, y: number) => [x - 1, y - 1],
      },
    ]

    return ranges.some(({ start, end, step }) => {
      for ([x, y] = start(); end(x, y); [x, y] = step(x, y)) {
        if (isUncovered(sensors, { x, y })) {
          return true
        }
      }
      return false
    })
  })

  return found ? { x, y } : undefined
}

function parseSensorData(lines: string[]): Sensor[] {
  return lines.map((line) => {
    const matcher = line.match(parser)
    if (matcher) {
      const sensor = {
        x: Number(matcher[1]),
        y: Number(matcher[2]),
      }
      const beacon = {
        x: Number(matcher[3]),
        y: Number(matcher[4]),
      }
      const distance = manHattanDistance(sensor, beacon)

      return {
        ...sensor,
        beacon,
        distance,
      } as Sensor
    }
    throw new Error(`Could not parse line: ${line}`)
  })
}

function solvePart1(lines: string[]): number {
  const sensors = parseSensorData(lines)
  const row = 2000000

  return findCoverageForRow(sensors, row)
}

function solvePart2(lines: string[]): number | undefined {
  const sensors = parseSensorData(lines)
  const signal = searchPerimeter(sensors, 0, 4000000)
  // const signal = findSignalNearby(map, 0, 20)

  return signal ? signal.x * 4000000 + signal.y : undefined
}

function run() {
  const input = process.argv.slice(2)[0]
  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)

  const solution1 = solvePart1(lines)
  console.log(`part1: ${solution1}`)

  const solution2 = solvePart2(lines)
  console.log(`part2: ${solution2}`)
}

run()
