import { readFileSync } from 'fs'

type Valve = {
  name: string
  pressure: number
  tunnels: Valve[]
  timeTo: { [key: string]: number }
}

const parser = /Valve (.+) has flow rate=(\d+); tunnels? leads? to valves? (.+)/

function parseValveData(lines: string[]): Valve[] {
  const tunnelMap: { [key: string]: string[] } = {}
  const valves = lines.map((line) => {
    const matcher = line.match(parser)
    if (matcher) {
      const name = matcher[1]
      const pressure = Number(matcher[2])
      tunnelMap[name] = matcher[3].split(', ')

      return {
        name,
        pressure,
        tunnels: [],
        timeTo: {},
      } as Valve
    }
    throw new Error(`Could not parse line: ${line}`)
  })

  valves.forEach((valve) => {
    valve.tunnels = tunnelMap[valve.name].map((name) => {
      const tunnel = valves.find((v) => v.name === name)
      if (tunnel) {
        return tunnel
      }
      throw new Error(`Could not find tunnel ${name}`)
    })
  })

  valves.forEach((valve) => {
    valves
      .filter((x) => x.pressure > 0)
      .forEach((otherValve) => {
        valve.timeTo[otherValve.name] = fastestTimeTo(valve, otherValve)
      })
  })

  return valves
}

function getPressure(valve: Valve, minutes: number): number {
  return valve.pressure * (minutes > 0 ? minutes : 0)
}

function fastestTimeTo(start: Valve, end: Valve): number {
  let count = 0
  const queue: [Valve, number][] = [[start, count]]
  const visited: string[] = [start.name]

  while (queue.length > 0) {
    const [valve, depth] = queue.shift() as [Valve, number]

    if (valve.name === end.name) {
      count = depth
      break
    }

    valve.tunnels
      .filter((v) => !visited.includes(v.name))
      .forEach((neighbor) => {
        queue.push([neighbor, depth + 1])
        visited.push(neighbor.name)
      })
  }

  return count
}

function maxPressureFor(
  pressuredValves: Valve[],
  valve: Valve,
  time: number,
  pressure = 0,
  callback?: (path: string[], pressure: number) => void,
  pathSeen: string[] = []
): number {
  if (time <= 1) {
    return pressure
  }

  if (pressuredValves.length === 0) {
    return pressure
  }

  return pressuredValves
    .map((valvePressured) => {
      let updatedTime = time
      let updatedPressure = pressure
      let updatedPath = pathSeen

      updatedTime = time - valve.timeTo[valvePressured.name] - 1
      updatedPressure = pressure + getPressure(valvePressured, updatedTime)
      updatedPath = [...pathSeen, valvePressured.name]

      callback?.(updatedPath, updatedPressure)

      return maxPressureFor(
        pressuredValves.filter((v) => v.name !== valvePressured.name),
        valvePressured,
        updatedTime,
        updatedPressure,
        callback,
        updatedPath
      )
    })
    .sort((a, b) => a - b)
    .pop() as number
}

function findBestTwoDisjointPaths(allPathScores: { [key: string]: number }): number {
  const orderedPaths = Object.entries(allPathScores)
    .sort(([, pressure1], [, pressure2]) => pressure2 - pressure1)
    .map(([key]) => key.split('-'))

  const best = orderedPaths.reduce((scores, path) => {
    orderedPaths.find((path2) => {
      if (path2.every((x) => !path.includes(x))) {
        const pathKey = path.join('-')
        const path2Key = path2.join('-')

        if (allPathScores[path2Key] && allPathScores[pathKey]) {
          scores.push(Math.max(allPathScores[path2Key] + allPathScores[pathKey]))
          return true
        }
      }
      return false
    })
    return scores
  }, [] as number[])

  return best.sort((a, b) => b - a).shift() as number
}

function findMaxPressure(valves: Valve[], minutes: number): number {
  const hasPressure = valves.filter((v) => v.pressure > 0)
  const valveAA = valves.find((v) => v.name === 'AA') as Valve
  return maxPressureFor(hasPressure, valveAA, minutes)
}

function findMaxPressureWithHelp(valves: Valve[], minutes: number): number {
  const cache: { [key: string]: number } = {}
  const caching = (path: string[], pressure: number) => {
    const key = path.sort().join('-')
    const getMemo = cache[key]
    cache[key] = getMemo ? Math.max(pressure, getMemo) : pressure
  }
  const hasPressure = valves.filter((v) => v.pressure > 0)
  const valveAA = valves.find((v) => v.name === 'AA') as Valve

  maxPressureFor(hasPressure, valveAA, minutes, 0, caching)
  return findBestTwoDisjointPaths(cache)
}

function solvePart1(lines: string[]): number {
  const valves = parseValveData(lines)
  return findMaxPressure(valves, 30)
}

function solvePart2(lines: string[]): number {
  const valves = parseValveData(lines)
  return findMaxPressureWithHelp(valves, 26)
}

function run() {
  const input = process.argv.slice(2)[0]
  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)

  const solution1 = solvePart1(lines)
  console.log(`part1: ${solution1}`)

  const solution2 = solvePart2(lines)
  console.log(`part2: ${solution2}`)
}

run()
