import { readFileSync } from 'fs'

const Mineral = {
  geode: 'geode',
  obsidian: 'obsidian',
  clay: 'clay',
  ore: 'ore',
} as const

type Minerals = keyof typeof Mineral

const allMinerals = Object.values(Mineral) as Minerals[]

type Inventory = {
  blueprint: number
  minerals: {
    [key in Minerals]: number
  }
  robots: {
    [key in Minerals]: number
  }
  toBuild?: Minerals
  atTime: number
}

type Blueprint = {
  number: number
  robots: {
    ore: {
      cost: {
        [key in Minerals]?: number
      }
    }
    clay: {
      cost: {
        [key in Minerals]?: number
      }
    }
    obsidian: {
      cost: {
        [key in Minerals]?: number
      }
    }
    geode: {
      cost: {
        [key in Minerals]?: number
      }
    }
  }
  maxRobots: {
    [key in Minerals]: number
  }
  time: number
}

const matchBlueprint =
  /Blueprint (\d+): .* (\d+) ore. .* (\d+) ore. .* (\d+) ore and (\d+) clay. .* (\d+) ore and (\d+) obsidian./

function parseBlueprint(lines: string[]): Blueprint[] {
  return lines.map((line) => {
    const match = line.match(matchBlueprint)
    if (match) {
      const [, number, oreOre, clayOre, obsidianOre, obsidianClay, geodeOre, geodeObsidian] =
        match.map((x) => Number(x))
      return {
        number,
        robots: {
          ore: { cost: { ore: oreOre } },
          clay: { cost: { ore: clayOre } },
          obsidian: {
            cost: { ore: obsidianOre, clay: obsidianClay },
          },
          geode: { cost: { ore: geodeOre, obsidian: geodeObsidian } },
        },
        maxRobots: {
          ore: Math.max(oreOre, clayOre, obsidianOre, geodeOre),
          clay: Math.max(0, obsidianClay),
          obsidian: Math.max(0, geodeObsidian),
          geode: Infinity,
        },
        time: 0,
      }
    } else {
      throw new Error(`Invalid cube: ${line}`)
    }
  })
}

function getRobotCostFor(blueprint: Blueprint, robot: Minerals, mineral: Minerals): number {
  return blueprint.robots[robot].cost[mineral] || 0
}

function build(inventory: Inventory, blueprint: Blueprint): Inventory {
  const minerals = { ...inventory.minerals }
  const robots = { ...inventory.robots }
  const time = inventory.atTime

  allMinerals.forEach((mineral) => {
    minerals[mineral] += robots[mineral]
    if (inventory.toBuild) {
      if (mineral === inventory.toBuild) {
        robots[mineral] += 1
      }
      minerals[mineral] -= getRobotCostFor(blueprint, inventory.toBuild, mineral)
    }
  })

  return { ...inventory, minerals, toBuild: undefined, robots, atTime: time - 1 }
}

function miningFor(inventory: Inventory, times: number): Inventory {
  const timeLeft = inventory.atTime - times

  if (timeLeft < 0) {
    throw new Error('Not enough time')
  }

  const minerals = { ...inventory.minerals }
  const robots = { ...inventory.robots }

  allMinerals.forEach((mineral) => {
    minerals[mineral] += robots[mineral] * times
  })

  return { ...inventory, minerals, robots, atTime: timeLeft }
}

function getMaxGeodes(inventory: Inventory): number {
  return inventory.minerals.geode + inventory.robots.geode * inventory.atTime
}

function estimateMaxGeodes(inventory: Inventory): number {
  const timeLeft = inventory.atTime - 1
  return getMaxGeodes(inventory) + (timeLeft * (2 + (timeLeft - 1))) / 2
}

function isFallingBehind(inventory: Inventory, maxGeodeInventory: Inventory): boolean {
  return estimateMaxGeodes(inventory) < getMaxGeodes(maxGeodeInventory)
}

function timeToBuildNextRobot(inventory: Inventory, blueprint: Blueprint, robot: Minerals): number {
  const mineralsNeeded = Object.keys(blueprint.robots[robot].cost) as Minerals[]
  const robotsAvailable = mineralsNeeded.filter((mineral) => inventory.robots[mineral] > 0)

  if (robotsAvailable.length !== mineralsNeeded.length) {
    return Infinity
  }

  const mineralsWanted = robotsAvailable.reduce((needs, mineral) => {
    needs[mineral] = getRobotCostFor(blueprint, robot, mineral)
    return needs
  }, {} as { [key in Minerals]: number })

  const waitTime = mineralsNeeded.reduce((maxTime, mineral) => {
    const needed = Math.max(0, mineralsWanted[mineral] - inventory.minerals[mineral])
    const time = needed === 0 ? 0 : Math.ceil(needed / inventory.robots[mineral])
    return Math.max(maxTime, time)
  }, 0)

  return waitTime
}

function getNextRobots(inventory: Inventory, blueprint: Blueprint): Inventory[] {
  const buildRobotInventory: Inventory[] = []

  allMinerals.forEach((robot) => {
    if (inventory.robots[robot] < blueprint.maxRobots[robot]) {
      const waitTime = timeToBuildNextRobot(inventory, blueprint, robot)
      if (inventory.atTime > waitTime) {
        const nextInventory = miningFor(inventory, waitTime)
        buildRobotInventory.push({ ...nextInventory, toBuild: robot })
      }
    }
  })

  return buildRobotInventory
}

function getMaxGeodesMined(inventory: Inventory, blueprint: Blueprint, time: number): Inventory {
  const visited: string[] = []
  const queue: Inventory[] = [{ ...inventory, atTime: time }]
  let maxGeodesInventory: Inventory = inventory

  while (queue.length) {
    const inventoryAtTime = queue.pop() as Inventory
    const time = inventoryAtTime.atTime
    const minerals = inventoryAtTime.minerals
    const robots = inventoryAtTime.robots

    if (getMaxGeodes(inventoryAtTime) > getMaxGeodes(maxGeodesInventory)) {
      maxGeodesInventory = miningFor(inventoryAtTime, time)
    }

    if (time === 0) {
      continue
    }

    const key = [...Object.values(minerals), time, ...Object.values(robots)].join('.')

    if (!visited.includes(key)) {
      visited.push(key)

      const combinations = getNextRobots(inventoryAtTime, blueprint)
        .map((inventory) => build(inventory, blueprint))
        .filter((inventory) => !isFallingBehind(inventory, maxGeodesInventory))
        .sort((a) => (a.toBuild === Mineral.geode ? -1 : 0))

      queue.push(...combinations)
    }
  }

  return maxGeodesInventory
}

function maxGeodesForBluePrint(blueprint: Blueprint, time: number): Inventory {
  const inventory: Inventory = {
    blueprint: blueprint.number,
    minerals: {
      ore: 0,
      clay: 0,
      obsidian: 0,
      geode: 0,
    },
    robots: {
      ore: 1,
      clay: 0,
      obsidian: 0,
      geode: 0,
    },
    atTime: 0,
  }

  return getMaxGeodesMined(inventory, blueprint, time)
}

function solvePart1(lines: string[]): number {
  const blueprints = parseBlueprint(lines)
  const inventories = blueprints.map((blueprint) => maxGeodesForBluePrint(blueprint, 24))
  const qualities = inventories.map((inventory) => inventory.minerals.geode * inventory.blueprint)
  return qualities.reduce((a, b) => a + b, 0)
}

function solvePart2(lines: string[]): number {
  const blueprints = parseBlueprint(lines).slice(0, 3)
  const inventories = blueprints.map((blueprint) => maxGeodesForBluePrint(blueprint, 32))
  return inventories.reduce((a, b) => a * b.minerals.geode, 1)
}

function run() {
  const input = process.argv.slice(2)[0]
  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)

  const solution1 = solvePart1(lines)
  console.log(`part1: ${solution1}`)

  const solution2 = solvePart2(lines)
  console.log(`part2: ${solution2}`)
}

run()
