import { FileHandle, open } from 'fs/promises'

const HandSignals = {
  rock: 'rock',
  paper: 'paper',
  scissor: 'scissor',
} as const

const GameOutcomes = {
  win: 'win',
  draw: 'draw',
  lose: 'lose',
} as const

const Hand = {
  A: HandSignals.rock,
  B: HandSignals.paper,
  C: HandSignals.scissor,
} as const

const GameOutcomeEncrypted = {
  X: GameOutcomes.lose,
  Y: GameOutcomes.draw,
  Z: GameOutcomes.win,
} as const

type HandSignal = keyof typeof HandSignals
type OtherPlayerHand = keyof typeof Hand
type GameOutcome = keyof typeof GameOutcomes
type GameEncryptedOutcome = keyof typeof GameOutcomeEncrypted

const YourHandScore: { [key in HandSignal]: number } = {
  rock: 1,
  paper: 2,
  scissor: 3,
}

const GameScore: { [key in GameOutcome]: number } = {
  win: 6,
  draw: 3,
  lose: 0,
}

function playOneGame(theirHand: HandSignal, yourHand: HandSignal): GameOutcome {
  if (theirHand === yourHand) {
    return GameOutcomes.draw
  }

  if (theirHand === HandSignals.rock) {
    return yourHand === HandSignals.paper ? GameOutcomes.win : GameOutcomes.lose
  }

  if (theirHand === HandSignals.scissor) {
    return yourHand === HandSignals.rock ? GameOutcomes.win : GameOutcomes.lose
  }

  if (theirHand === HandSignals.paper) {
    return yourHand === HandSignals.scissor ? GameOutcomes.win : GameOutcomes.lose
  }

  throw new Error('Invalid hand')
}

function findHandGivenOutcome(theirHand: HandSignal, outcome: GameOutcome): HandSignal | undefined {
  return Object.values(HandSignals).find((signal) => playOneGame(theirHand, signal) === outcome)
}

function scoreGame(theirHand: HandSignal, outcome: GameOutcome): number {
  const yourHand = findHandGivenOutcome(theirHand, outcome)

  if (yourHand) {
    const yourPlayScore = YourHandScore[yourHand]
    const yourOutcomeScore = GameScore[outcome]

    return yourPlayScore + yourOutcomeScore
  }

  throw new Error('Invalid game')
}

async function scoreAllGames(file: FileHandle): Promise<number> {
  let totalScore = 0

  for await (const line of file.readLines()) {
    const [theirHand, outcome] = line.split(' ') as [OtherPlayerHand, GameEncryptedOutcome]
    totalScore += scoreGame(Hand[theirHand], GameOutcomeEncrypted[outcome])
  }

  return totalScore
}

async function run() {
  const input = process.argv.slice(2)[0]
  const fd = await open(input)
  const totalScore = await scoreAllGames(fd)

  console.log(`solution: ${totalScore}`)
}

run()
