import { readFileSync } from 'fs'

type Operands = '-' | '+' | '*' | '/'

type MonkeyJobNumber = {
  type: 'number'
  value: number
}

type MonkeyJobVariable = {
  type: 'variable'
  value: string
}

type MonkeyJobMath = {
  left: string
  right: string
  type: 'math'
  operand: Operands
}

type MonkeyJob = MonkeyJobMath | MonkeyJobNumber | MonkeyJobVariable

type MonkeyJobs = {
  [monkeyName: string]: MonkeyJob
}

type MonkeyEquation =
  | [number, Operands, string | MonkeyEquation]
  | [string | MonkeyEquation, Operands, number]

const jobMatch = /^(.+): (.+)$/
const jobTypeNumberMatch = /^\d+$/
const jobTypeMathMatch = /^(.+) ([-+*/]) (.+)$/

function parseJobs(lines: string[]): MonkeyJobs {
  return lines.reduce((jobs, line) => {
    const matchJob = line.match(jobMatch)
    if (matchJob) {
      const [, monkeyName, job] = matchJob

      if (jobTypeNumberMatch.test(job)) {
        jobs[monkeyName] = {
          value: parseInt(job, 10),
          type: 'number',
        } as MonkeyJobNumber
      } else if (jobTypeMathMatch.test(job)) {
        const jobMath = job.match(jobTypeMathMatch)
        if (jobMath) {
          const [, leftMonkeyName, operand, rightMonkeyName] = jobMath

          jobs[monkeyName] = {
            left: leftMonkeyName,
            operand,
            right: rightMonkeyName,
            type: 'math',
          } as MonkeyJobMath
        } else {
          throw new Error(`Invalid math job: ${job}`)
        }
      }
    } else {
      throw new Error(`Invalid job: ${line}`)
    }
    return jobs
  }, {} as MonkeyJobs)
}

function solveJobFor(monkey: string, monkeyJobs: MonkeyJobs): number {
  const monkeyJob = monkeyJobs[monkey]

  if (monkeyJob) {
    if (monkeyJob.type === 'number') {
      return monkeyJob.value
    } else if (monkeyJob.type === 'math') {
      const leftValue = solveJobFor(monkeyJob.left, monkeyJobs)
      const rightValue = solveJobFor(monkeyJob.right, monkeyJobs)
      return calculate(leftValue, monkeyJob.operand, rightValue)
    }
  }

  throw new Error(`no job for monkey: ${monkey}`)
}

function calculate(left: number, operand: Operands, right: number): number {
  switch (operand) {
    case '+':
      return left + right
    case '-':
      return left - right
    case '*':
      return left * right
    case '/':
      return left / right
    default:
      throw new Error(`invalid operand: ${operand}`)
  }
}

function getAlgorithmFor(
  monkeyJob: MonkeyJob,
  monkeyJobs: MonkeyJobs
): string | number | MonkeyEquation {
  if (monkeyJob.type === 'math') {
    const { operand, left, right } = monkeyJob
    const leftAlgo = getAlgorithmFor(monkeyJobs[left], monkeyJobs)
    const rightAlgo = getAlgorithmFor(monkeyJobs[right], monkeyJobs)

    if (typeof leftAlgo === 'number') {
      if (typeof rightAlgo === 'number') {
        return calculate(leftAlgo, operand, rightAlgo)
      }

      return [leftAlgo, operand, rightAlgo]
    } else if (typeof rightAlgo === 'number') {
      return [leftAlgo, operand, rightAlgo]
    }
  } else {
    return monkeyJob.value
  }

  throw new Error(`too many dependencies for: ${monkeyJob}`)
}

function solve(equation: string | MonkeyEquation, equals: number): number {
  const [left, operand, right] = equation

  if (typeof equation === 'string') {
    return equals
  }

  // left + right = x -> right = x - left -> left = x - right
  // left - right = x -> right = left - x -> left = x + right
  // left * right = x -> right = x / left -> left = x / right
  // left / right = x -> right = left / x -> left = x * right

  if (typeof left === 'number' && typeof right !== 'number') {
    switch (operand) {
      case '+':
        return solve(right, equals - left)
      case '-':
        return solve(right, left - equals)
      case '*':
        return solve(right, equals / left)
      default:
        return solve(right, left / equals)
    }
  } else if (typeof right === 'number' && typeof left !== 'number') {
    switch (operand) {
      case '+':
        return solve(left, equals - right)
      case '-':
        return solve(left, equals + right)
      case '*':
        return solve(left, equals / right)
      default:
        return solve(left, equals * right)
    }
  }

  throw new Error(`invalid equation: ${equation}`)
}

function solvePart1(lines: string[]): number {
  const monkeyJobs = parseJobs(lines)
  return solveJobFor('root', monkeyJobs)
}

function solvePart2(lines: string[]): number {
  const monkeyJobs = parseJobs(lines)
  const root = monkeyJobs['root']

  monkeyJobs['humn'] = { type: 'variable', value: 'x' }

  if (root.type === 'math') {
    const rootAlgo = getAlgorithmFor({ ...root, operand: '-' }, monkeyJobs)

    if (Array.isArray(rootAlgo)) {
      return solve(rootAlgo, 0)
    }
  }

  throw new Error('algo failed to be solved')
}

function run() {
  const input = process.argv.slice(2)[0]
  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)

  const solution1 = solvePart1(lines)
  console.log(`part1: ${solution1}`)

  const solution2 = solvePart2(lines)
  console.log(`part2: ${solution2}`)
}

run()
