import { readFileSync } from 'fs'

const TrailSpaceType = {
  OPEN: '.',
  WALL: '#',
  EMPTY: ' ',
  UP: '^',
  DOWN: 'v',
  LEFT: '<',
  RIGHT: '>',
} as const

const InstructionTurn = {
  CLOCKWISE: 'R',
  COUNTER_CLOCKWISE: 'L',
} as const

const TrailDirection = {
  UP: 'UP',
  DOWN: 'DOWN',
  LEFT: 'LEFT',
  RIGHT: 'RIGHT',
} as const

type TrailSpaceTypes = typeof TrailSpaceType[keyof typeof TrailSpaceType]
type TrailDirections = typeof TrailDirection[keyof typeof TrailDirection]
type TrailInstructionTurns = typeof InstructionTurn[keyof typeof InstructionTurn]

type TrailSpace = {
  type: TrailSpaceTypes
  point: TrailPoint
}

type TrailPoint = {
  x: number
  y: number
}

type TrailPosition = {
  space: TrailSpace
  facing: TrailDirections
}

type Trail = {
  start: TrailPosition
  spaces: { [coordinates: string]: TrailSpace }
  height: number
  width: number
  cube: {
    size: number
    indexes: [number, number][]
    edges: TrailSpace[]
  }
}

type TrailInstruction = {
  turn?: TrailInstructionTurns
  distance: number
}

type TrailCubeFace = {
  index: [number, number]
  folds: TrailDirections[]
  rotation?: {
    turn: TrailInstructionTurns
    times: number
  }
}

const TrailCompass = [
  TrailDirection.RIGHT,
  TrailDirection.DOWN,
  TrailDirection.LEFT,
  TrailDirection.UP,
]

const matchTrailInstruction = /\d+(R|L)?/g

function getNeighbors(trail: Trail, { x, y }: TrailPoint) {
  const up = y < 0 ? trail.height - 1 : y - 1
  const down = y > trail.height - 1 ? 0 : y + 1
  const left = x < 0 ? trail.width - 1 : x - 1
  const right = x > trail.width - 1 ? 0 : x + 1

  return {
    [TrailDirection.DOWN]: { point: { x, y: down }, dir: TrailDirection.DOWN },
    [TrailDirection.UP]: { point: { x, y: up }, dir: TrailDirection.UP },
    [TrailDirection.LEFT]: { point: { x: left, y }, dir: TrailDirection.LEFT },
    [TrailDirection.RIGHT]: { point: { x: right, y }, dir: TrailDirection.RIGHT },
  }
}

function parseTrail(lines: string[]): Trail {
  const start = { point: { x: 0, y: 0 }, type: TrailSpaceType.EMPTY, neighbors: {} }
  const trail: Trail = {
    start: {
      space: start,
      facing: TrailDirection.RIGHT,
    },
    spaces: {},
    height: lines.length,
    width: lines.reduce((width, line) => Math.max(width, line.length), 0),
    cube: {
      size: lines.length,
      indexes: [],
      edges: [],
    },
  }
  const column = { min: trail.height, max: 0 }

  lines.forEach((line, y) => {
    const spaces = line.split('') as TrailSpaceTypes[]
    const row = { min: trail.width, max: 0 }

    spaces.forEach((space, x) => {
      if (TrailSpaceType.EMPTY !== space) {
        const trailSpace: TrailSpace = {
          point: { x, y },
          type: space,
        }

        if (trail.start.space.type === TrailSpaceType.EMPTY) {
          trail.start.space = trailSpace
        }

        setSpace(trail, trailSpace)

        row.min = Math.min(row.min, x)
        row.max = Math.max(row.max, x)
        column.min = Math.min(column.min, y)
        column.max = Math.max(column.max, y)
      }
    })

    trail.cube.size = Math.min(trail.cube.size, row.max - row.min + 1)
  })

  if (trail.cube.size !== Math.min(trail.cube.size, column.max - column.min + 1)) {
    throw new Error('Not a cube')
  }

  trail.cube.edges = Object.values(trail.spaces).filter((space) =>
    Object.values(getNeighbors(trail, space.point)).some(
      (neighbor) => !getSpace(trail, neighbor.point)
    )
  )

  trail.cube.indexes = getFaceIndexes(trail.cube.edges, trail.cube.size)

  return trail
}

const arrayEquals = <T>(one: T[], two: T[]): boolean =>
  one.length === two.length && one.every((value, i) => value === two[i])

function getTurnsBeforeRotation(dirs: TrailDirections[]) {
  return dirs.reduce((foundIndex, _, index) => {
    if (dirs[index + 1] !== undefined) {
      return dirs[index] === dirs[index + 1] ? foundIndex : index
    }
    return foundIndex
  }, 0)
}

function getRotation(dirs: TrailDirections[]) {
  return dirs.reduce((turn, _, index) => {
    if (index + 1 < dirs.length) {
      const first = TrailCompass.indexOf(dirs[index])
      const second = TrailCompass.indexOf(dirs[index + 1])
      const isCounterClockwise = first - second === 1 || second - first === 3
      const isClockwise = second - first === 1 || first - second === 3

      if (isClockwise && isCounterClockwise) {
        return turn
      } else if (turn === InstructionTurn.COUNTER_CLOCKWISE && isClockwise) {
        return null
      } else if (turn === InstructionTurn.CLOCKWISE && isCounterClockwise) {
        return null
      } else if (isCounterClockwise) {
        return InstructionTurn.COUNTER_CLOCKWISE
      } else if (isClockwise) {
        return InstructionTurn.CLOCKWISE
      }
    }
    return turn
  }, null as TrailInstructionTurns | null)
}

function isFlipped(dirs: TrailDirections[]) {
  return dirs.some((dir, index) => {
    const nextFolds = dirs.slice(index, index + 3)
    return nextFolds.length === 3 && nextFolds.every((fold) => fold === dir)
  })
}

function findNextFaceFor(
  trail: Trail,
  index: [number, number],
  dir: TrailDirections
): TrailCubeFace {
  const queue: TrailCubeFace[] = [{ index, folds: [] }]
  const visited: [number, number][] = []

  while (queue.length > 0) {
    const faceMap = queue.shift() as TrailCubeFace
    const lastFold = faceMap.folds.slice(-1)[0]
    const turnDir = getRotation(faceMap.folds)
    const turnTimes = getTurnsBeforeRotation(faceMap.folds) + 1

    if (
      lastFold === dir ||
      (isFlipped(faceMap.folds) && getOppositeDirection(lastFold) === dir) ||
      visited.length === 5
    ) {
      if (turnDir) {
        faceMap.rotation = {
          turn: turnDir,
          times: turnTimes,
        }
      }
      return faceMap
    }

    const x = faceMap.index[0]
    const y = faceMap.index[1]

    Object.values(getNeighbors(trail, { x, y }))
      .map((neighbor) => {
        const turnFold = turnDir
          ? turnDirectionTimes(neighbor.dir, turnDir, turnTimes)
          : neighbor.dir
        return {
          index: [neighbor.point.x, neighbor.point.y],
          folds: [...faceMap.folds, turnFold],
        } as TrailCubeFace
      })
      .filter((neighbor) => trail.cube.indexes.find((index) => arrayEquals(index, neighbor.index)))
      .filter((neighbor) => !visited.find((visit) => arrayEquals(visit, neighbor.index)))
      .forEach((neighbor) => {
        queue.push(neighbor)
      })

    visited.push(faceMap.index)
  }

  throw new Error(`Could not find face for ${index} and ${dir}`)
}

function getOppositeTurn(turn: TrailInstructionTurns): TrailInstructionTurns {
  return turn === InstructionTurn.COUNTER_CLOCKWISE
    ? InstructionTurn.CLOCKWISE
    : InstructionTurn.COUNTER_CLOCKWISE
}

function getOppositeDirection(dir: TrailDirections): TrailDirections {
  const index = TrailCompass.indexOf(dir)
  const oppositeIndex =
    index + 2 > TrailCompass.length - 1 ? (index + 2) % TrailCompass.length : index + 2
  return TrailCompass[oppositeIndex]
}

function getFaceIndexes(edges: TrailSpace[], size: number): [number, number][] {
  return edges
    .map((edge) => getFaceIndex(edge, size))
    .reduce((indexes, faceIndex) => {
      if (!indexes.find((index) => arrayEquals(index, faceIndex))) {
        indexes.push(faceIndex)
      }
      return indexes
    }, [] as [number, number][])
}

function findFoldedEdgeFor(
  trail: Trail,
  fromEdge: TrailSpace,
  fromDirection: TrailDirections
): TrailSpace {
  const fromIndex = getFaceIndex(fromEdge, trail.cube.size)
  const toFaceMap = findNextFaceFor(trail, fromIndex, fromDirection)
  if (!toFaceMap) {
    throw new Error(`Could not find edge for ${fromEdge}`)
  }
  const oppositeDir = getOppositeDirection(fromDirection)
  const oppositeTurn = getOppositeTurn(toFaceMap.rotation?.turn || InstructionTurn.CLOCKWISE)
  const toDirection = toFaceMap.rotation?.turn
    ? turnDirectionTimes(oppositeDir, oppositeTurn, toFaceMap.rotation.times)
    : fromDirection
  const fromMod = getMod(fromEdge, fromDirection, trail.cube.size)
  const toEdge = trail.cube.edges
    .filter((edge) => arrayEquals(getFaceIndex(edge, trail.cube.size), toFaceMap.index))
    .filter((edge) => !getSpace(trail, getNeighbors(trail, edge.point)[toDirection].point))
    .find((edge) => {
      return (
        fromMod ===
        getMod(
          edge,
          toDirection,
          trail.cube.size,
          toFaceMap.rotation?.turn,
          toFaceMap.rotation?.times
        )
      )
    })

  if (!toEdge) {
    throw new Error(`Could not connect edge for ${JSON.stringify(fromEdge.point)}`)
  }

  return toEdge
}

function parseInstructions(line: string): TrailInstruction[] {
  const matches = line.match(matchTrailInstruction) || []
  return matches.map((match) => {
    if (/^\d+$/.test(match)) {
      return { distance: parseInt(match) }
    } else {
      return {
        turn: match.slice(-1) as TrailInstructionTurns,
        distance: parseInt(match.slice(0, -1)),
      }
    }
  })
}

function turnDirectionTimes(
  facing: TrailDirections,
  dir: TrailInstructionTurns,
  times: number
): TrailDirections {
  let next = facing
  for (let i = 0; i < times; i++) {
    next = turnDirection(next, dir)
  }
  return next
}

function turnDirection(facing: TrailDirections, dir: TrailInstructionTurns): TrailDirections {
  const compass = TrailCompass
  const index = compass.indexOf(facing)
  const oppositeIndex = InstructionTurn.CLOCKWISE === dir ? index + 1 : index - 1

  if (oppositeIndex < 0) {
    return compass[compass.length - 1]
  }
  return compass[oppositeIndex % compass.length]
}

function positionMove(
  trail: Trail,
  position: TrailPosition,
  instruction: TrailInstruction,
  asCube = false
): TrailPosition {
  let dir = position.facing
  let at = position.space
  let next: TrailSpace | undefined = at

  for (let i = 0; i < instruction.distance; i++) {
    at.type = TrailSpaceType[dir]

    let neighbor = getNeighbors(trail, at.point)[dir]
    next = getSpace(trail, neighbor.point)

    if (!next) {
      if (asCube) {
        next = findFoldedEdgeFor(trail, at, dir)
      } else {
        while (!next) {
          neighbor = getNeighbors(trail, neighbor.point)[dir]
          next = getSpace(trail, neighbor.point)
        }
      }
    }

    if (!next) {
      throw new Error(`No next space for ${at.point.y},${at.point.y} in direction ${dir}`)
    }

    if (next.type !== TrailSpaceType.WALL) {
      if (asCube) {
        const atIndex = getFaceIndex(at, trail.cube.size)
        const nextIndex = getFaceIndex(next, trail.cube.size)
        if (!arrayEquals(atIndex, nextIndex)) {
          const face = findNextFaceFor(trail, atIndex, dir)
          if (face.rotation) {
            dir = turnDirectionTimes(dir, getOppositeTurn(face.rotation.turn), face.rotation.times)
          }
        }
      }
      at = next
    }
  }

  return {
    space: at,
    facing: instruction.turn ? turnDirection(dir, instruction.turn) : dir,
  }
}

function getSpace(trail: Trail, point: TrailPoint): TrailSpace | undefined {
  return trail.spaces[`${point.x},${point.y}`]
}

function setSpace(trail: Trail, space: TrailSpace) {
  trail.spaces[`${space.point.x},${space.point.y}`] = space
}

function getFaceIndex(space: TrailSpace, size: number): [number, number] {
  const x = Math.ceil((space.point.x + 1) / size) - 1
  const y = Math.ceil((space.point.y + 1) / size) - 1

  return [x, y]
}

function getMod(
  space: TrailSpace,
  dir: TrailDirections,
  size: number,
  turn?: TrailInstructionTurns,
  rotations = 0
): number {
  switch (dir) {
    case TrailDirection.UP:
    case TrailDirection.DOWN:
      if (
        rotations === 0 ||
        (turn && turn === InstructionTurn.CLOCKWISE && (rotations === 4 || rotations === 1)) ||
        (turn && turn === InstructionTurn.COUNTER_CLOCKWISE && (rotations === 3 || rotations === 4))
      ) {
        return space.point.x % size
      } else {
        return size - (space.point.x % size) - 1
      }
    case TrailDirection.LEFT:
    case TrailDirection.RIGHT:
      if (
        rotations === 0 ||
        (turn && turn === InstructionTurn.CLOCKWISE && (rotations === 4 || rotations === 3)) ||
        (turn && turn === InstructionTurn.COUNTER_CLOCKWISE && (rotations === 1 || rotations === 4))
      ) {
        return space.point.y % size
      } else {
        return size - (space.point.y % size) - 1
      }
  }
}

function printTrail(trail: Trail): void {
  for (let y = 0; y < trail.height; y++) {
    const row: string[] = []
    for (let x = 0; x < trail.width; x++) {
      row.push(getSpace(trail, { x, y })?.type || ' ')
    }
    console.log(row.join(''))
  }
}

function parseInput(lines: string[]): {
  trail: Trail
  instructions: TrailInstruction[]
} {
  if (!lines.length) {
    throw new Error('could not parse input')
  }

  const instructionLine = lines.slice(-1)[0]
  const instructions = parseInstructions(instructionLine)
  const trail = parseTrail(lines.slice(0, -2))

  return { trail, instructions }
}

function password(destination: TrailPosition): number {
  const row = 1000 * (destination.space.point.y + 1)
  const column = 4 * (destination.space.point.x + 1)
  const face = TrailCompass.indexOf(destination.facing)

  return row + column + face
}

function solvePart1(lines: string[], debug = false): number {
  const { trail, instructions } = parseInput(lines)
  const positions: TrailPosition[] = [trail.start]

  instructions.forEach((instruction) => {
    const position = positionMove(trail, positions.slice(-1)[0], instruction)
    positions.push(position)
  })

  if (debug) {
    printTrail(trail)
  }

  return password(positions.slice(-1)[0])
}

function solvePart2(lines: string[], debug = false): number {
  const { trail, instructions } = parseInput(lines)
  const positions: TrailPosition[] = [trail.start]

  instructions.forEach((instruction) => {
    const position = positionMove(trail, positions.slice(-1)[0], instruction, true)
    positions.push(position)
  })

  if (debug) {
    printTrail(trail)
  }

  return password(positions.slice(-1)[0])
}

function run() {
  const input = process.argv.slice(2)[0]
  const debug = process.env['DEBUG'] ? true : false

  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)

  const solution1 = solvePart1(lines, debug)
  console.log(`part1: ${solution1}`)

  const solution2 = solvePart2(lines, debug)
  console.log(`part2: ${solution2}`)
}

run()
