import { readFileSync } from 'fs'

const DIRECTION = {
  NORTH: 'N',
  SOUTH: 'S',
  EAST: 'E',
  WEST: 'W',
} as const

const SPACE = {
  EMPTY: '.',
  ELF: '#',
} as const

type Directions = typeof DIRECTION[keyof typeof DIRECTION]
type Spaces = typeof SPACE[keyof typeof SPACE]
type CraterPoint = { x: number; y: number }
type Elf = {
  id: number
  at: CraterPoint
  lastProposal: CraterPoint
}
type Crater = {
  elves: { [coordinates: string]: Elf }
  proposals: { [coordinates: string]: number }
  width: { from: number; to: number }
  height: { from: number; to: number }
}

function getPointKey(point: CraterPoint): string {
  return `${point.x},${point.y}`
}

function getCraterSpace(crater: Crater, point: CraterPoint): Spaces {
  return crater.elves[getPointKey(point)] ? SPACE.ELF : SPACE.EMPTY
}

function isIsolated(crater: Crater, elf: Elf): boolean {
  const { x, y } = elf.at

  const around = [
    { x, y: y - 1 },
    { x, y: y + 1 },

    { x: x + 1, y: y - 1 },
    { x: x + 1, y: y + 1 },
    { x: x + 1, y: y },

    { x: x - 1, y: y - 1 },
    { x: x - 1, y: y + 1 },
    { x: x - 1, y: y },
  ]

  return around.every((point) => getCraterSpace(crater, point) !== SPACE.ELF)
}

function canPropose(crater: Crater, elf: Elf, direction: Directions): boolean {
  const { x, y } = elf.at

  const proposals = {
    [DIRECTION.NORTH]: [
      { x, y: y - 1 },
      { x: x + 1, y: y - 1 },
      { x: x - 1, y: y - 1 },
    ],
    [DIRECTION.SOUTH]: [
      { x, y: y + 1 },
      { x: x + 1, y: y + 1 },
      { x: x - 1, y: y + 1 },
    ],
    [DIRECTION.EAST]: [
      { x: x + 1, y: y },
      { x: x + 1, y: y + 1 },
      { x: x + 1, y: y - 1 },
    ],
    [DIRECTION.WEST]: [
      { x: x - 1, y: y },
      { x: x - 1, y: y + 1 },
      { x: x - 1, y: y - 1 },
    ],
  }

  return proposals[direction].every((point) => getCraterSpace(crater, point) !== SPACE.ELF)
}

function makeProposal(crater: Crater, elf: Elf, order: Directions[]): CraterPoint {
  for (const direction of order) {
    if (!isIsolated(crater, elf) && canPropose(crater, elf, direction)) {
      switch (direction) {
        case DIRECTION.NORTH:
          return { x: elf.at.x, y: elf.at.y - 1 }
        case DIRECTION.SOUTH:
          return { x: elf.at.x, y: elf.at.y + 1 }
        case DIRECTION.EAST:
          return { x: elf.at.x + 1, y: elf.at.y }
        case DIRECTION.WEST:
          return { x: elf.at.x - 1, y: elf.at.y }
      }
    }
  }

  return { ...elf.at }
}

function numberOfProposals(crater: Crater, elf: Elf): number {
  return crater.proposals[getPointKey(elf.lastProposal)]
}

function setProposal(crater: Crater, elf: Elf, proposal: CraterPoint) {
  const key = getPointKey(proposal)
  const proposals = crater.proposals[key]

  crater.proposals[key] = proposals ? proposals + 1 : 1
  elf.lastProposal = proposal
}

function makeProposals(crater: Crater, order: Directions[]) {
  Object.values(crater.elves).forEach((elf) => {
    const proposal = makeProposal(crater, elf, order)
    setProposal(crater, elf, proposal)
  })
}

function settleProposals(crater: Crater) {
  const height = { from: 0, to: 0 }
  const width = { from: 0, to: 0 }
  const elves = { ...crater.elves }

  crater.elves = {}

  Object.values(elves).forEach((elf, index) => {
    if (numberOfProposals(crater, elf) === 1) {
      elf.at = { ...elf.lastProposal }
    } else {
      if (!numberOfProposals(crater, elf)) {
        console.log(elf)
      }
    }

    crater.elves[getPointKey(elf.at)] = elf

    if (index === 0) {
      height.to = elf.at.y
      height.from = elf.at.y

      width.to = elf.at.x
      width.from = elf.at.x
    } else {
      height.to = Math.max(height.to, elf.at.y)
      height.from = Math.min(height.from, elf.at.y)

      width.to = Math.max(width.to, elf.at.x)
      width.from = Math.min(width.from, elf.at.x)
    }
  })

  crater.height = height
  crater.width = width
  crater.proposals = {}
}

function parseInput(lines: string[]): Crater {
  const crater: Crater = {
    elves: {},
    proposals: {},
    width: { from: 0, to: lines[0].length - 1 },
    height: { from: 0, to: lines.length - 1 },
  }

  const spaces = lines.map((line) => line.split('') as Spaces[])

  spaces.flatMap((row, i) =>
    row.forEach((space, j) => {
      if (space === SPACE.ELF) {
        const point = { x: j, y: i }
        const key = getPointKey(point)
        const elf = {
          at: point,
          lastProposal: point,
          id: i * crater.height.to + j,
        } as Elf

        crater.elves[key] = elf
      }
    })
  )

  return crater
}

function printCrater(crater: Crater) {
  for (let y = crater.height.from; y <= crater.height.to; y++) {
    const row = [] as Spaces[]
    for (let x = crater.width.from; x <= crater.width.to; x++) {
      row.push(getCraterSpace(crater, { x, y }))
    }
    console.log(row.join(''))
  }
  console.log('\n')
}

function countCraterEmptySpaces(crater: Crater): number {
  const height = Math.abs(crater.height.to - crater.height.from)
  const width = Math.abs(crater.width.to - crater.width.from)

  return (height + 1) * (width + 1) - Object.keys(crater.elves).length
}

function compareElves(elves1: Crater['elves'], elves2: Crater['elves']): boolean {
  return !Object.keys(elves1).some((key) => elves2[key] === undefined)
}

function solvePart1(lines: string[], debug = false): number {
  const crater = parseInput(lines)
  const order = [DIRECTION.NORTH, DIRECTION.SOUTH, DIRECTION.WEST, DIRECTION.EAST]

  if (debug) {
    printCrater(crater)
  }

  for (let round = 0; round < 10; round++) {
    makeProposals(crater, order)
    settleProposals(crater)

    if (debug) {
      printCrater(crater)
    }

    order.push(order.shift() as Directions)
  }

  return countCraterEmptySpaces(crater)
}

function solvePart2(lines: string[], debug = false): number {
  const crater = parseInput(lines)
  const order = [DIRECTION.NORTH, DIRECTION.SOUTH, DIRECTION.WEST, DIRECTION.EAST]
  let rounds = 0

  if (debug) {
    printCrater(crater)
  }

  while ((rounds += 1)) {
    const prevElves = { ...crater.elves }

    makeProposals(crater, order)
    settleProposals(crater)

    if (debug) {
      printCrater(crater)
    }

    if (compareElves(prevElves, crater.elves)) {
      return rounds
    }

    order.push(order.shift() as Directions)
  }

  return 0
}

function run() {
  const input = process.argv.slice(2)[0]
  const debug = process.env.DEBUG ? true : false

  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)

  const solution1 = solvePart1(lines, debug)
  console.log(`part1: ${solution1}`)

  const solution2 = solvePart2(lines, debug)
  console.log(`part2: ${solution2}`)
}

run()
