import { readFileSync } from 'fs'

const BlizzardMarker = {
  BLIZZARD_UP: '^',
  BLIZZARD_DOWN: 'v',
  BLIZZARD_LEFT: '<',
  BLIZZARD_RIGHT: '>',
} as const

const ValleySpace = {
  ...BlizzardMarker,
  WALL: '#',
  EMPTY: '.',
  ME: 'E',
  BLIZARD_MANY: '*',
} as const

type Point = { x: number; y: number }
type ValleySpaces = typeof ValleySpace[keyof typeof ValleySpace]
type BlizzardMarkers = typeof ValleySpace[keyof typeof BlizzardMarker]

type Blizzard = {
  marker: BlizzardMarkers
  start: Point
}

type Valley = {
  width: number
  height: number
  start: Point
  exit: Point
  blizzards: Blizzard[]
  spaceSnapshots: { [key: string]: Record<string, ValleySpaces> }
}

function pointToKey(point: Point): string {
  return `${point.x},${point.y}`
}

function findFastestStepsBetween(valley: Valley, start: Point, exit: Point, timeStart = 0): number {
  const queue = [{ point: start, time: timeStart }] as {
    point: Point
    time: number
  }[]
  const visited = {} as { [time: string]: { [key: string]: boolean } }

  let fastest = Infinity

  while (queue.length) {
    const { point, time } = queue.shift() as typeof queue[number]
    const key = pointToKey(point)
    const cycle = (valley.width - 2) * (valley.height - 2)
    const cycleTime = (time + 1) % cycle

    if (time >= fastest) {
      continue
    }

    if (visited[key]?.[cycleTime]) {
      continue
    }

    if (point.x === exit.x && point.y === exit.y) {
      fastest = time
      continue
    }

    const nextMoves = [
      { x: point.x - 1, y: point.y },
      { x: point.x + 1, y: point.y },

      { x: point.x, y: point.y - 1 },
      { x: point.x, y: point.y + 1 },

      { x: point.x, y: point.y },
    ]

    if (!visited[key]) {
      visited[key] = {}
    }

    visited[key][cycleTime] = true

    nextMoves
      .filter((neighbor) => getSpaceAtPoint(neighbor, valley, time + 1) === ValleySpace.EMPTY)
      .forEach((neighbor) => {
        queue.push({ point: neighbor, time: time + 1 })
      })
  }

  return fastest
}

function getSpaceAtPoint(point: Point, valley: Valley, time: number): ValleySpaces {
  const snapshot = getSpacePositions(valley, time)
  const space = snapshot[`${point.x},${point.y}`]

  if (point.x === valley.start.x && point.y === valley.start.y) {
    return ValleySpace.EMPTY
  } else if (point.x === valley.exit.x && point.y === valley.exit.y) {
    return ValleySpace.EMPTY
  } else if (point.x <= 0 || point.x >= valley.width - 1) {
    if (space) {
      throw new Error(`blizzard is in a wall at ${point.x},${point.y}`)
    }

    return ValleySpace.WALL
  } else if (point.y <= 0 || point.y >= valley.height - 1) {
    if (space) {
      throw new Error(`blizzard is in a wall at ${point.x},${point.y}`)
    }

    return ValleySpace.WALL
  } else if (space) {
    return space
  } else {
    return ValleySpace.EMPTY
  }
}

function blizzardPositionAtTime(blizzard: Blizzard, valley: Valley, time: number): Point {
  const start = blizzard.start
  const maxWidth = valley.width - 2
  const maxHeight = valley.height - 2

  let x = start.x
  let y = start.y

  switch (blizzard.marker) {
    case ValleySpace.BLIZZARD_UP:
      y = start.y - time
      break
    case ValleySpace.BLIZZARD_DOWN:
      y = start.y + time
      break
    case ValleySpace.BLIZZARD_LEFT:
      x = start.x - time
      break
    case ValleySpace.BLIZZARD_RIGHT:
      x = start.x + time
      break
  }

  if (y < 1) {
    y = maxHeight - Math.abs(y % maxHeight)
  } else if (y > maxHeight) {
    y = 1 + ((y - 1) % maxHeight)
  }

  if (x < 1) {
    x = maxWidth - Math.abs(x % maxWidth)
  } else if (x > maxWidth) {
    x = 1 + ((x - 1) % maxWidth)
  }

  return { x, y }
}

function parseInput(lines: string[]): Valley {
  const valley: Valley = {
    width: lines[0].length,
    height: lines.length,
    start: { x: 0, y: 0 },
    exit: { x: 0, y: 0 },
    blizzards: [],
    spaceSnapshots: {},
  }

  lines.forEach((line, y) => {
    line.split('').forEach((space, x) => {
      if (space === ValleySpace.EMPTY) {
        if (y === 0) {
          valley.start = { x, y }
        } else if (y === lines.length - 1) {
          valley.exit = { x, y }
        }
      } else if (Object.values(BlizzardMarker).includes(space as BlizzardMarkers)) {
        valley.blizzards.push({ marker: space, start: { x, y } } as Blizzard)
      }
    })
  })

  return valley
}

function getSpacePositions(valley: Valley, time: number): Record<string, ValleySpaces> {
  const timeCycle = time % ((valley.width - 2) * (valley.height - 2))
  const snapshot = valley.spaceSnapshots[timeCycle]

  if (snapshot) {
    return snapshot
  } else {
    const snapshots = valley.spaceSnapshots

    snapshots[timeCycle] = valley.blizzards.reduce((pointMap, blizzard) => {
      const point = blizzardPositionAtTime(blizzard, valley, time)
      const key = `${point.x},${point.y}`
      pointMap[key] = pointMap[key] ? ValleySpace.BLIZARD_MANY : blizzard.marker
      return pointMap
    }, {} as Record<string, ValleySpaces>)

    return snapshots[timeCycle]
  }
}

function printValley(valley: Valley, me: Point, time: number) {
  const line: string[] = []
  for (let y = 0; y < valley.height; y++) {
    for (let x = 0; x < valley.width; x++) {
      const point = { x, y }
      if (point.x === me.x && point.y === me.y) {
        line.push(ValleySpace.ME)
      } else {
        const space = getSpaceAtPoint(point, valley, time)

        line.push(space)
      }
    }
    line.push('\n')
  }

  console.log(line.join(''))
}

function solvePart1(lines: string[], debug = false): number {
  const valley = parseInput(lines)

  if (debug) {
    printValley(valley, valley.start, 0)
  }

  return findFastestStepsBetween(valley, valley.start, valley.exit, 0)
}

function solvePart2(lines: string[], debug = false): number {
  const valley = parseInput(lines)

  if (debug) {
    printValley(valley, valley.start, 0)
  }

  const timeToExit = findFastestStepsBetween(valley, valley.start, valley.exit, 0)
  const timeToSnack = findFastestStepsBetween(valley, valley.exit, valley.start, timeToExit)
  const timeToExitAgain = findFastestStepsBetween(valley, valley.start, valley.exit, timeToSnack)

  return timeToExitAgain
}

async function run() {
  const input = process.argv.slice(2)[0]
  const debug = process.env['DEBUG'] ? true : false

  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)

  const solution1 = solvePart1(lines, debug)
  console.log(`part1: ${solution1}`)

  const solution2 = solvePart2(lines, debug)
  console.log(`part2: ${solution2}`)
}

run()
