import { readFileSync } from 'fs'

const SnafuMap = {
  '1': 1,
  '2': 2,
  '=': -2,
  '-': -1,
  '0': 0,
} as const
const SnafuBase = 5

type Snafus = keyof typeof SnafuMap
type SnafuNumbers = typeof SnafuMap[Snafus]

const snafuMapKeys = Object.keys(SnafuMap) as Snafus[]
const snafusOrdered = snafuMapKeys.sort((a, b) => SnafuMap[a] - SnafuMap[b])

function parseInput(lines: string[]): Snafus[][] {
  return lines.map((line) => {
    const strings = line.split('') as Snafus[]
    if (strings.every((s) => s in SnafuMap)) {
      return strings
    }
    throw new Error(`Invalid snafu string: ${line}`)
  })
}

function getSnafuFromNumber(snafuNumber: SnafuNumbers): Snafus {
  const entry = Object.entries(SnafuMap).find(([, value]) => snafuNumber === value)

  if (entry) {
    return entry[0] as Snafus
  }

  throw new Error(`Invalid snafu lookup for ${snafuNumber}`)
}

function snafuPower(place: number, snafu: Snafus): number {
  return Math.pow(SnafuBase, place) * SnafuMap[snafu]
}

function snafuRange(places: number): number {
  let total = 0

  for (let i = 0; i < places; i++) {
    total += snafuPower(i, getSnafuFromNumber(2))
  }

  return total
}

function getSnafuStringAt(input: number, snafus: Snafus[], at: number): Snafus[] {
  const currentValue = snafuStringsToDecimal(snafus)
  const copySnafus = [...snafus]

  if (input - currentValue !== 0) {
    const snafu = snafusOrdered.find((findSnafu) => {
      const testSnafus = [...snafus]
      testSnafus[at] = findSnafu

      const testDecimal = snafuStringsToDecimal(testSnafus)
      const testHighRange = snafuRange(snafus.length - at - 1)

      return testDecimal + testHighRange >= input && testDecimal - testHighRange <= input
    })

    if (snafu === undefined) {
      return getSnafuStringAt(input, [...copySnafus, getSnafuFromNumber(0)], at)
    } else {
      copySnafus[at] = snafu
    }

    if (at === snafus.length - 1) {
      return copySnafus
    } else {
      return getSnafuStringAt(input, copySnafus, at + 1)
    }
  }

  while (at < snafus.length) {
    copySnafus[at] = getSnafuFromNumber(0)
    at += 1
  }

  return snafus
}

function decimalToSnafuStrings(input: number): Snafus[] {
  return getSnafuStringAt(input, [] as Snafus[], 0)
}

function snafuStringsToDecimal(digits: Snafus[]): number {
  return [...digits]
    .map((c) => SnafuMap[c])
    .reverse()
    .reduce((total, digit, i) => total + snafuPower(i, getSnafuFromNumber(digit)), 0 as number)
}

function solvePart1(lines: string[], debug = false): string {
  const decimals = parseInput(lines).map(snafuStringsToDecimal)
  const total = decimals.reduce((total, decimal) => total + decimal, 0)
  const snafuStrings = decimalToSnafuStrings(total)

  if (debug && total !== snafuStringsToDecimal(snafuStrings)) {
    throw new Error('snafu')
  }

  return snafuStrings.join('')
}

async function run() {
  const input = process.argv.slice(2)[0]
  const debug = process.env.DEBUG ? true : false

  const file = readFileSync(input, 'utf8')
  const lines = file.split('\n').slice(0, -1)

  const solution = solvePart1(lines, debug)
  console.log(`solution: ${solution}`)
}

run()
