import { FileHandle, open } from 'fs/promises'

type File = {
  name: string
  size: number
}

type FileSystem = {
  directories: FileSystem[]
  files: File[]
  parent?: FileSystem
  name: string
  size: number
}

const DISK_SPACE = 70000000
const DESIRED_FREE_SPACE = 30000000
const matchCommandChangeDir = (line: string) => line.match(/^\$ cd (.+)/)
const matchListFile = (line: string) => line.match(/^(\d+) (.+)/)

const rootFs: FileSystem = {
  directories: [],
  files: [],
  name: '/',
  size: 0,
}

function findDir(fs: FileSystem, dir: string): FileSystem | undefined {
  return fs.directories.find((d) => d.name === dir)
}

function appendDir(parentFs: FileSystem, dir: string): FileSystem {
  const fs: FileSystem = { name: dir, size: 0, files: [], parent: parentFs, directories: [] }
  parentFs.directories.push(fs)

  return fs
}

function maybeAppendDir(parentFs: FileSystem, dir: string): FileSystem {
  return findDir(parentFs, dir) || appendDir(parentFs, dir)
}

function findFile(fs: FileSystem, name: string): File | undefined {
  return fs.files.find((d) => d.name === name)
}

function appendFile(parentFs: FileSystem, name: string, size: number): void {
  parentFs.files.push({ name, size })
}

function maybeAppendFile(parentFs: FileSystem, name: string, size: number): void {
  if (!findFile(parentFs, name)) {
    appendFile(parentFs, name, size)
  }
}

function walkAndCacheSize(fs: FileSystem): void {
  const allFileSizes = fs.files.reduce((acc, dir) => acc + dir.size, 0)
  const allDirSizes = fs.directories
    .map((dir) => {
      walkAndCacheSize(dir)
      return dir.size
    })
    .reduce((acc, dir) => acc + dir, 0)
  fs.size += allDirSizes + allFileSizes
}

function sortDirectories(fs: FileSystem): FileSystem[] {
  return [
    fs,
    ...fs.directories.flatMap((dir) => {
      return sortDirectories(dir).sort((a, b) => a.size - b.size)
    }),
  ].sort((a, b) => a.size - b.size)
}

async function solve(file: FileHandle): Promise<void> {
  let currentFs = rootFs

  for await (const line of file.readLines()) {
    const matchChangeDir = matchCommandChangeDir(line)
    const matchFile = matchListFile(line)

    if (matchChangeDir) {
      const dir = matchChangeDir[1]
      if (dir !== rootFs.name) {
        if (dir === '..' && currentFs.parent) {
          currentFs = currentFs.parent
        } else {
          currentFs = maybeAppendDir(currentFs, dir)
        }
      }
    } else if (matchFile) {
      const file = matchFile[2] as string
      const size = matchFile[1] as string
      maybeAppendFile(currentFs, file, +size)
    }
  }

  walkAndCacheSize(rootFs)
  const freeSpace = DESIRED_FREE_SPACE - (DISK_SPACE - rootFs.size)
  const possibleDeletions = sortDirectories(rootFs).filter((fs) => fs.size > freeSpace)

  if (process.env.DEBUG) {
    console.log(possibleDeletions)
  }

  console.log(`solution: ${possibleDeletions[0].size}`)
}

async function run() {
  const input = process.argv.slice(2)[0]
  const fd = await open(input)
  await solve(fd)
}

run()
