import { FileHandle, open } from 'fs/promises'

type Position = { x: number; y: number }
type Instruction = { direction: 'R' | 'U' | 'L' | 'D'; distance: number }

function moveHead(head: Position, direction: Instruction['direction']) {
  if (direction === 'R') {
    head.x += 1
  } else if (direction === 'L') {
    head.x -= 1
  } else if (direction === 'U') {
    head.y += 1
  } else if (direction === 'D') {
    head.y -= 1
  }
}

function areTouching(knotOne: Position, knotTwo: Position) {
  return Math.abs(knotOne.x - knotTwo.x) <= 1 && Math.abs(knotOne.y - knotTwo.y) <= 1
}

function followKnot(head: Position, tail: Position) {
  const { x: hx, y: hy } = head
  const { x: tx, y: ty } = tail

  const shouldMoveVertically = Math.abs(ty - hy) > 1
  const shouldMoveHorizontally = Math.abs(tx - hx) > 1
  const shouldMoveDiagonally = !areTouching(head, tail) && tx !== hx && ty !== hy

  if (shouldMoveVertically || shouldMoveDiagonally) {
    tail.y += ty < hy ? 1 : -1
  }

  if (shouldMoveHorizontally || shouldMoveDiagonally) {
    tail.x += tx < hx ? 1 : -1
  }
}

function move(rope: Position[], instruction: Instruction, tailPositions: Set<string>) {
  const head = rope[0]
  const tail = rope[rope.length - 1]
  let remainingMoves = instruction.distance

  while (remainingMoves > 0) {
    moveHead(head, instruction.direction)

    rope.forEach((knot, index) => {
      if (index !== 0) {
        followKnot(rope[index - 1], knot)
      }
    })

    // console.log(instruction, head, tail)
    tailPositions.add(`${tail.x},${tail.y}`)
    remainingMoves -= 1
  }
}

function parseInstruction(line: string): Instruction {
  const match = line.match(/^([RULD]) (\d+)$/)
  if (match) {
    return { direction: match[1] as Instruction['direction'], distance: +match[2] }
  } else {
    throw new Error(`invalid instruction: ${line}`)
  }
}

async function solve(file: FileHandle): Promise<void> {
  const tailPositions = new Set<string>()
  const rope = Array(10)
    .fill({})
    .map(() => ({ x: 0, y: 0 } as Position))

  for await (const line of file.readLines()) {
    const instruction = parseInstruction(line)
    move(rope, instruction, tailPositions)
  }

  console.log(`solution: ${tailPositions.size}`)
}

async function run() {
  const input = process.argv.slice(2)[0]
  const fd = await open(input)
  await solve(fd)
}

run()
